﻿Imports System.Globalization
Imports System.Text
Imports System.Text.RegularExpressions

Module Program


    Dim CI As CultureInfo
    Dim nDecimals As Int32
    Dim line As String = ""
    Dim value As Double, unitsName As String
    Sub Main()
        DspHelp()
        Do
            line = Menu()
        Loop While Not CnslFns.functionKeys.Keys.Contains(line)
    End Sub
    Function Menu()
        Dim iQuestion As Int32 = 1
        Do
            line = ""
            Select Case iQuestion
                Case 1 ' Q1
                    line = "en-US"
                    line = Cnsl.ReadLine("Please, enter your 'CultureInfo': ", True, False, line)
                    If CnslFns.IsNotAFunctKey(line) Then
                        Try
                            CI = New CultureInfo(line)
                            iQuestion += 1
                        Catch ex As Exception
                            DspErrorMsg("Please enter a valid CultureInfo (en-US, fr_FR, es-ES, ...")
                        End Try
                    End If
                Case 2 ' Q2
                    nDecimals = 15
                    line = Cnsl.ReadLine(vbCrLf + "How many decimals (0-15)?: " + vbCrLf, True, False, nDecimals.ToString)
                    If CnslFns.IsNotAFunctKey(line) Then
                        If Int32.TryParse(line, nDecimals) AndAlso 0 <= nDecimals AndAlso nDecimals <= 15 Then
                            iQuestion += 1
                        Else
                            DspErrorMsg("Please enter an integer between 0 and 15.")
                        End If
                    End If
                Case 3 ' Q3
                    line = Cnsl.ReadLine("What is the units value?: ", True, True)
                    If CnslFns.IsNotAFunctKey(line) Then
                        If Double.TryParse(line, CI, value) Then
                            iQuestion += 1
                            Continue Do
                        Else
                            DspErrorMsg("Please enter a valid numeric value.")
                        End If
                    End If
                Case 4 ' Q4
                    line = Cnsl.ReadLine("...and the units name?: ", True, True, "", 0) ' allow 0 characters
                    If CnslFns.IsNotAFunctKey(line) Then
                        unitsName = line
                        DspValueAndUnitsNameEngNotation(value, unitsName)
                        iQuestion = 3
                    End If
            End Select
            Do While Not CnslFns.IsNotAFunctKey(line)
                Select Case line
                    Case CnslFns.functionKeys("DspHelp")
                        DspHelp() : line = ""
                    Case CnslFns.functionKeys("AskAgain")
                        iQuestion = 1 : line = ""
                    Case CnslFns.functionKeys("Examples")
                        line = Examples()
                        iQuestion = 2
                    Case CnslFns.functionKeys("CLS")
                        Console.Clear()
                        Return ""
                    Case CnslFns.functionKeys("Quit")
                        Return line
                End Select
            Loop
        Loop
        Return line
    End Function
    Function Examples() As String
        Dim line As String = ""
        Try
            Dim dicExamples = New Dictionary(Of Double, String) _
            From {{123456.789, "V"}, {0.123456, "V"},
                 {-0.000012, "H"}, {12.345 * 10 ^ -12, "F"}}
            For i = 0 To dicExamples.Count - 1
                Console.WriteLine($"{i + 1} {dicExamples.Keys(i)} {dicExamples.Values(i).ToString(CI)}")
            Next
            Dim n As Int32
            Do
                line = Cnsl.ReadLine("Select an example #: ", True, False)
                Select Case line
                    Case CnslFns.functionKeys("DspHelp") ' pressed F1, display help '
                        DspHelp() : line = ""
                        Continue Do
                    Case CnslFns.functionKeys("AskAgain") ' pressed F2, repeat questions '
                        Exit Do
                    Case CnslFns.functionKeys("Examples") ' pressed F3, examples '
                        Exit Do
                    Case CnslFns.functionKeys("CLS") ' user pressed F4, CLear Screen '
                        Console.Clear()
                        Exit Do
                    Case CnslFns.functionKeys("Quit") ' user pressed F5, Quit '
                        Exit Do
                End Select
                If Int32.TryParse(line, n) Then
                    If 0 < n AndAlso n <= dicExamples.Count Then
                        value = dicExamples.Keys(n - 1)
                        unitsName = dicExamples.Values(n - 1)
                        DspValueAndUnitsNameEngNotation(value, unitsName)
                        Exit Do ' Selected Option was OK
                    End If
                End If
                DspErrorMsg("Please enter a valid example number.")
            Loop
        Catch ex As Exception
            DspErrorMsg($"Error {ex.Message}")
            Cnsl.ReadLine("", True, minLength:=0)
            line = CnslFns.functionKeys("Quit")
        End Try
        Return line
    End Function
    Sub DspValueAndUnitsNameEngNotation(value As Double, unitsName As String)
        Dim sOutput As String = ToEngineeringNotation(value,
                                                    unitsName,
                                                    nDecimals,
                                                    CI)
        Console.WriteLine("")
        Console.WriteLine("Units value: " + value.ToString(CI))
        Console.WriteLine("Units name.: " + unitsName)
        Console.ForegroundColor = ConsoleColor.Green
        Console.WriteLine("Output.....: " + sOutput + vbCrLf)
        Console.ForegroundColor = ConsoleColor.White
    End Sub
    Sub DspErrorMsg(msg As String)
        Console.ForegroundColor = ConsoleColor.Red
        Console.WriteLine(msg)
        Console.ForegroundColor = ConsoleColor.White
    End Sub
    Sub DspHelp()
        Console.WriteLine("F1 = help / F2 = start / F3 = examples / F4 = CLS / F5 = Quit")
    End Sub

    Function ToEngineeringNotation(value As Double,
                                       Optional unitName As String = "",
                                       Optional nDecimals As Int32 = 15,
                                       Optional CI As Globalization.CultureInfo = Nothing) As String

        Static incPrefixes() As String = {"k", "M", "G", "T", "P", "E", "Z", "Y"}
        Static decPrefixes() As String = {"m", "µ", "n", "p", "f", "a", "z", "y"}

        If CI Is Nothing Then CI = New Globalization.CultureInfo("en-US")
        Dim sgn As Int32 = Math.Sign(value)
        value = Math.Abs(value)
        Dim gteOne As Boolean = IIf(value >= 1, True, False)
        Dim s As String = value.ToString(CI).ToUpper
        If s.IndexOf("E") = -1 Then s += "E+00"
        Dim vs() As String = Split(s, "E")
        Dim exp As Int32 = Int32.Parse(vs(1), CI)
        Dim m As Match = Regex.Match(vs(0), Regex.Escape(CI.NumberFormat.CurrencyDecimalSeparator))
        If m.Success Then
            vs(0) = vs(0).Remove(m.Index, 1)
            exp -= vs(0).Length - m.Index
        End If
        value = Decimal.Parse(vs(0))

        Dim suffix As String = ""
        If gteOne OrElse value = 0 Then
            value *= 10 ^ exp
            Dim i As Int32 = -1
            If value >= 1000 Then
                i = Math.Log10(value) \ 3
                value /= 10 ^ (3 * i)
            End If
            If i > -1 Then suffix = incPrefixes(i - 1)
        Else
            Dim i As Int32 = (-exp + 2) \ 3
            value *= 10 ^ (3 * i + exp)
            If value >= 1000 Then
                Dim thnds As Int32 = Math.Log10(value) \ 3
                value /= 10 ^ (3 * thnds)
                i -= thnds
            End If
            suffix = decPrefixes(i - 1)
        End If
        suffix += unitName
        Dim sFormat As String = "{0:#0." + New String("#", nDecimals) + "} {1}"
        Return String.Format(CI, sFormat, sgn * value, suffix)
    End Function

End Module
